# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
from typing import Dict

from taurus_datajob_api import ApiException
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import UserCodeError
from vdk.plugin.meta_jobs.meta import IDataJobExecutor
from vdk.plugin.meta_jobs.meta import TrackableJob
from vdk.plugin.meta_jobs.remote_data_job import JobStatus

log = logging.getLogger(__name__)

ACTIVE_JOB_STATUSES = [JobStatus.SUBMITTED.value, JobStatus.RUNNING.value]


class TrackingDataJobExecutor:
    def __init__(self, executor: IDataJobExecutor):
        self._executor = executor
        self._jobs_cache: Dict[str, TrackableJob] = dict()

    def register_job(self, job: TrackableJob):
        if job.job_name in self._jobs_cache:
            log.warning(
                f"Job with name {job.job_name} aleady exists. Details: {self._jobs_cache[job.job_name]}. "
                f"Will overwrite it."
            )
        self._jobs_cache[job.job_name] = job

    def start_job(self, job_name: str) -> None:
        """
        :param job_name: the job to start and track
        """
        job = self._get_job(job_name)
        job.start_attempt += 1
        execution_id = self._executor.start_job(job.job_name, job.team_name)
        log.info(f"Starting new data job execution with id {execution_id}")
        self._jobs_cache[job.job_name] = job
        job.execution_id = execution_id
        job.status = JobStatus.SUBMITTED.value
        job.details = self._executor.details_job(
            job.job_name, job.team_name, job.execution_id
        )
        log.info(
            f"Started data job {job_name}:\n{self._get_printable_details(job.details)}"
        )

    def finalize_job(self, job_name):
        job = self._get_job(job_name)
        details = self._executor.details_job(
            job.job_name, job.team_name, job.execution_id
        )
        job.details = details
        log.info(
            f"Finished data job {job_name}:\n"
            f"start_time: {details['start_time']}\n"
            f"end_time: {details['end_time']}\n"
            f"status: {details['status']}\n"
            f"message: {details['message']}"
        )
        if job.status != JobStatus.SUCCEEDED.value and job.fail_meta_job_on_error:
            raise UserCodeError(
                ErrorMessage(
                    "",
                    "Meta Job failed due to a data job failure.",
                    f"Data Job {job_name} failed. See details: {details}",
                    "The rest of the jobs in the meta job will not be started "
                    "and the meta data job will fail.",
                    "Investigate the error in the job or re-try again.",
                )
            )

    @staticmethod
    def _get_printable_details(details):
        del details["deployment"]
        return json.dumps(details, default=lambda o: str(o), indent=2)

    def _get_job(self, job_name) -> TrackableJob:
        job: TrackableJob = self._jobs_cache.get(job_name)
        if job is None:
            raise IndexError(
                f"The job {job_name} has not been registered. Use register_job first. "
            )
        return job

    def status(self, job_name: str) -> str:
        job = self._get_job(job_name)
        if job.status in ACTIVE_JOB_STATUSES:
            job.status = self._executor.status_job(
                job.job_name, job.team_name, job.execution_id
            )
        log.debug(f"Job status: {job}")
        return job.status

    def get_finished_job_names(self):
        finalized_jobs = []
        # TODO: optimize
        # Do not call the status every time (use TTL caching)
        # Do not call all status at the same time - stagger them in time
        for job in self._jobs_cache.values():
            if job.status is not None and job.status in ACTIVE_JOB_STATUSES:
                job.status = self.status(job.job_name)

        for job in self._jobs_cache.values():
            if job.status is not None and job.status not in ACTIVE_JOB_STATUSES:
                finalized_jobs.append(job.job_name)
        return finalized_jobs

    def get_all_jobs(self):
        return list(self._jobs_cache.values())



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import abc
from dataclasses import dataclass

from vdk.plugin.meta_jobs.api import meta_job


class IDataJobExecutor(abc.ABC):
    @abc.abstractmethod
    def start_job(self, job_name: str, team_name: str):
        """
        Start an execution of a data job and returns its execution id
        :param job_name:
        :param team_name:
        :return: execution id
        """
        pass

    @abc.abstractmethod
    def status_job(self, job_name: str, team_name: str, execution_id: str):
        """
        Get the current status of a data job execution
        :param job_name:
        :param team_name:
        :param execution_id:
        :return: status in string as defined by Control Service API
        """
        pass

    @abc.abstractmethod
    def details_job(self, job_name: str, team_name: str, execution_id: str) -> dict:
        """
        Get the current status of a data job execution
        :param job_name:
        :param team_name:
        :param execution_id:
        :return: status in string as defined by Control Service API
        """
        pass


@dataclass
class TrackableJob(meta_job.SingleJob):
    status: str = None
    execution_id: str = None
    details: dict = None
    start_attempt = 0



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from abc import ABC
from typing import Dict

from vdk.internal.control.configuration.vdk_config import VDKConfig
from vdk.plugin.meta_jobs.meta import IDataJobExecutor
from vdk.plugin.meta_jobs.remote_data_job import RemoteDataJob


class RemoteDataJobExecutor(IDataJobExecutor):
    def start_job(self, job_name: str, team_name: str):
        vdk_cfg = VDKConfig()
        job = RemoteDataJob(job_name, team_name, vdk_cfg.control_service_rest_api_url)
        return job.start_job_execution()
        # catch error on 409

    def status_job(self, job_name: str, team_name: str, execution_id: str) -> str:
        vdk_cfg = VDKConfig()
        job = RemoteDataJob(job_name, team_name, vdk_cfg.control_service_rest_api_url)
        status = job.get_job_execution_status(execution_id)
        return status.status

    def details_job(self, job_name: str, team_name: str, execution_id: str) -> dict:
        vdk_cfg = VDKConfig()
        job = RemoteDataJob(job_name, team_name, vdk_cfg.control_service_rest_api_url)
        status = job.get_job_execution_status(execution_id)
        return status.to_dict()



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import json
import logging
from typing import Dict
from typing import List
from typing import Optional

from vdk.plugin.meta_jobs.api.meta_job import IMetaJobInput
from vdk.plugin.meta_jobs.api.meta_job import MetaJob
from vdk.plugin.meta_jobs.meta_dag import MetaJobsDag

TEAM_NAME: Optional[str] = None

log = logging.getLogger(__name__)


def get_json(obj):
    return json.loads(json.dumps(obj, default=lambda o: o.__dict__))


class MetaJobInput(IMetaJobInput):
    def run_meta_job(self, jobs: List[Dict]):
        dag = MetaJobsDag(TEAM_NAME)
        dag.build_dag(jobs)
        dag.execute_dag()
        log.info(f"Meta job summary:\n{dag}")



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import time
from collections import deque
from dataclasses import dataclass
from random import randint
from random import random
from typing import Any


@dataclass
class _QueueEntry:
    element: Any
    ready_time: float


class TimeBasedQueue:
    """
    Use it to enqueue elements until element is ready to be dequeued.
    """

    def __init__(
        self,
        min_ready_time_seconds=10,
        randomize_delay_seconds=10,
        dequeue_timeout_seconds=5,
    ):
        """

        :param min_ready_time_seconds: the minimum time in seconds for element to be ready
        :param randomize_delay_seconds: added delay in seconds randomly between [0 , random_delay_seconds]
        :param dequeue_timeout_seconds: dequeue method blocks maximum that many seconds before returning.
        """
        self._elements: deque[_QueueEntry] = deque()
        self._min_ready_time_seconds = min_ready_time_seconds
        self._randomize_delay_seconds = randomize_delay_seconds
        self._dequeue_timeout_seconds = dequeue_timeout_seconds

    def enqueue(self, element):
        ready_time = self._min_ready_time_seconds + randint(
            0, self._randomize_delay_seconds
        )
        self._elements.append(_QueueEntry(element, time.time() + ready_time))

    def dequeue(self):
        """
        Dequeues the next ready element. Element is ready if their ready time (duration) has passed.
        Otherwise it will wait up to dequeue_timeout_seconds and return None if no element is ready.
        If the queue is empty it returns immediately with None.
        """
        if len(self._elements) > 0:
            entry = self._elements[0]
            if entry.ready_time <= time.time():
                return self._elements.popleft().element
            else:
                self.__wait_for_entry_to_be_ready(entry.ready_time)
                if entry.ready_time <= time.time():
                    return self._elements.popleft().element
        return None

    def size(self):
        return len(self._elements)

    def __wait_for_entry_to_be_ready(self, ready_time: float):
        sleep_time_seconds = ready_time - time.time()
        sleep_time_seconds = min(sleep_time_seconds, self._dequeue_timeout_seconds)
        time.sleep(sleep_time_seconds)
